/*
 * (C) 2012, Roland Baudin <roland65@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

// Plugin includes
#include "USMFilter.h"
#include "gauss_iir.h"

// Bibble includes
#include "PluginPipeSettings.h"
#include "PluginTile.h"
#include "PluginImage.h"
#include "PluginImageSettings.h"
#include "PluginOptionList.h"

// C++ includes
#include <cmath>
#include <iostream>

// Qt includes
#include <QDebug>


using namespace std;


// Number of neighboring tiles needed by the plugin (don't set it to 0!)
#define NEIGHBORTILES 1


// Get the value of the option that indicates if the filter must be run or not 
bool USMFilter::needsToRunLayer(const PluginImageSettings &options, const PluginPipeSettings &settings, PluginOptionList  &layerOptions) const
{
    Q_UNUSED(options);
    Q_UNUSED(settings);

    bool okay;

    // Get the value of the enable option
    bool val = layerOptions.getBool(Enable, m_groupId, okay);

    return val;
}


// Indicate if processing will be slow (to allocate more buffers?)
bool USMFilter::isSlow(const PluginImageSettings &options, const PluginPipeSettings &settings) const
{
    Q_UNUSED(options);
    Q_UNUSED(settings);
    
    return true; // Processing is slow
}


// Indicate if processing needs to use neighbor tiles
bool USMFilter::isLargeRadius(const PluginImageSettings &options, const PluginPipeSettings &settings) const
{
    Q_UNUSED(options);
    Q_UNUSED(settings);
    
    return true; // We use neighbor tiles
}


// Indicate if we need the original tile to process a layer
bool USMFilter::needsOriginalImage()
{
    return true; // We need the original tile
}


// Indicate if there are pixels that are generated from scratch
bool USMFilter::isPixelSource(const PluginImageSettings &options, const PluginPipeSettings &settings) const
{
    Q_UNUSED(options);
    Q_UNUSED(settings);

    return false; // We don't generate pixels from scratch
}


// Indicate if the image is geometrically transformed
QTransform USMFilter::transform(const PluginImageSettings &options, const PluginPipeSettings &settings) const
{
    Q_UNUSED(options);
    Q_UNUSED(settings);

    return QTransform(); // We don't warp the image so return an identity transform
}


// Indicate if the image size is changed
QSize USMFilter::size(const PluginImageSettings &options, const PluginPipeSettings &settings, const QSize &startSize) const
{
    Q_UNUSED(options);
    Q_UNUSED(settings);

    return startSize;	// We don't change the size of the image
}


// Prerequisites
QList<PluginDependency*> USMFilter::prerequisites(const PluginImageSettings &options, const PluginPipeSettings &settings, PluginTile &tile) const
{
    Q_UNUSED(options);
    
    int srcStep = m_hub->step(this) - 1;
    
    QRect tileRect = QRect(QPoint(tile.x(),tile.y()), settings.tileSize());

    int tileExtension = NEIGHBORTILES*settings.tileSize().width();

    // Make rectangle bigger by NEIGHBORTILES * 256 pixels in each direction
    QRect srcRect = tileRect.adjusted(-tileExtension, -tileExtension, tileExtension, tileExtension);
    return settings.tilesIn(srcRect, srcStep);
}


// Perform the filtering
void USMFilter::ProcessBuffer(float *fimg[3], int width, int height, float zoomLevel, const PluginImageSettings &options, const PluginPipeSettings &settings, PluginTile &tile, PluginOptionList &layerOptions, int layerPos) const
{
    Q_UNUSED(options);
    Q_UNUSED(settings);
    Q_UNUSED(tile);
    Q_UNUSED(layerPos);

    // Here, the filter is run just before the ASP sharpening filter
    // Thus fimg[0][], fimg[1][] and fimg[2][] are RGB colors in [0,1]
    // We first perform a RGB -> LAB conversion and the sharpening is done on the L channel
    // The USM filter is therefore applied to fimg[0][]

	// RGB -> LAB conversion
	for (int i = 0;i<(width*height);i++)
	{
	    rgb2lab(&fimg[0][i],&fimg[1][i],&fimg[2][i]);
	}
    
    // Make a copy of the L channel   
    float *org = new float[width*height];
    memcpy(org,fimg[0],width*height*sizeof(float));
    bool okay;

    // Read options
    bool  enable = layerOptions.getBool(Enable, m_groupId, okay);
    float amount = layerOptions.getDouble(Amount, m_groupId, okay)/100.0f;
    float radius = layerOptions.getDouble(Radius, m_groupId, okay)*zoomLevel;
    float threshold = layerOptions.getInt(Threshold, m_groupId, okay);
        
    if (enable)
    {       
        // Blur
        GAUSS_IIR(fimg[0],width,height,radius);

        // Blend
        for (int i = 0;i<(width*height);i++)
        {
            // Threshold
            float diff = (org[i] - fimg[0][i]);            
            diff = (diff>0.0f)?max(diff-threshold,0.0f):min(diff+threshold,0.0f);
            
            // USM
            diff *= amount;
            fimg[0][i] = org[i] + diff; // We don't limit L
        }
    }

    // Free memory
    delete[] org;

	// LAB -> RGB conversion
	for (int i = 0;i<(width*height);i++)
	{
	    lab2rgb(&fimg[0][i],&fimg[1][i],&fimg[2][i]);
		
		// Limit RGB data to [0,1]
		fimg[0][i] = CLIPF(fimg[0][i],0.0f,1.0f);
		fimg[1][i] = CLIPF(fimg[1][i],0.0f,1.0f);
		fimg[2][i] = CLIPF(fimg[2][i],0.0f,1.0f);
	}
}


// Compute additional radius depending on the zoom level
int USMFilter::additionalRadius(PluginOptionList  &layerOptions,float zoomLevel) const
{
    bool okay;
    bool  enable = layerOptions.getBool(Enable, m_groupId, okay);
    float radius = layerOptions.getDouble(Radius, m_groupId, okay)*zoomLevel;
    float effRadius = 1.0f + (enable?(radius/2.0):0.0f);

    return effRadius*4;
}


// Perform the processing
void USMFilter::runLayer(const PluginImageSettings &options, const PluginPipeSettings &settings, PluginTile &tile, PluginOptionList  &layerOptions, int layerPos) const
{
    const bool runsOnLayer = (layerPos != 0);

    // Destination image
    PluginImage *targetImage;

    //	Layer 0 output should always go into tile.image(), otherwise output should go into tile.tmpImage()
    if (runsOnLayer)
        targetImage = tile.tmpImage();
    else
        targetImage = tile.image();

    int srcStep = m_hub->step(this) - 1;

	// Rectangle extension = NEIGHBORTILES * 256
    int tileExtension = NEIGHBORTILES*settings.tileSize().width();
  
 	// Number of tiles in array
	const int numTiles = 2*NEIGHBORTILES + 1;

    PluginImage *srcImages[numTiles][numTiles];
    uint16      *srcData[numTiles][numTiles];
    int          rowWords[numTiles][numTiles];
    int          activeWidth[numTiles][numTiles];
    int          activeHeight[numTiles][numTiles];

	// Tile dimensions (256 pixels)
    int tileWidth  = settings.tileSize().width();
    int tileHeight = settings.tileSize().height();

    // Size of the whole image at the current mip level
    QSize size = settings.size(srcStep);
    int width  = size.width();
    int height = size.height();

    // Number of existing tiles in image
    int tilesW = width/tileWidth +  ((width%tileWidth)?1:0);
    int tilesH = height/tileHeight + ((height%tileHeight)?1:0);

    // Get an array of the tile images needed
    int col, row;
    for (row=0; row<numTiles; row++)
        for (col=0; col<numTiles; col++)
        {
            int ro = tile.row() - NEIGHBORTILES + row;
            int co = tile.col() - NEIGHBORTILES + col;
            if ((ro <0 || co < 0)  || (ro >= tilesH) || co >= tilesW   )
            {
                srcImages[col][row] = NULL;
                srcData[col][row] = NULL;
            }
            else
            {
                PluginTile *cTile = (dynamic_cast<PluginTile*>(settings.tileAt(co,ro, srcStep)));
                srcImages[col][row] = cTile->image();
                
                if (srcImages[col][row] != NULL)
                	srcData[col][row] = srcImages[col][row]->data16();
                else
                	srcData[col][row] = NULL;
                
                rowWords[col][row] = srcImages[col][row]->rowWords();
                activeWidth[col][row] = cTile->activeSize().width();
                activeHeight[col][row]= cTile->activeSize().height();
            }
        }

    PluginImage *image = (dynamic_cast<PluginTile*>(settings.tileAt(tile.col(),tile.row(), srcStep)))->image();

	// Increment to get to the next pixel
    int pw = image->planeWords();
    int np = image->nextPixel();

	// Paddings
    int pad = settings.tilePad();
    int padh = pad;
    int padw = pad;
    
    // Read size extension
    const float zoomLevel = pow(0.5f,settings.mip());
    float radius = additionalRadius(layerOptions,zoomLevel);

    int bufferWidth  = tileWidth + 2*tileExtension;
    int bufferHeight = tileHeight + 2*tileExtension;

 	// Amount of unused pixels in the neighbor tiles
 	const int shrinkBuffers = max(int(tileWidth-(radius*2 + 2*pad)),0);
    int totalPixels = (bufferWidth-2*shrinkBuffers)*(bufferHeight-2*shrinkBuffers);

    // Allocate the fimg buffer
    float *fimg[3];
    for (int i = 0; i < 3; i++)
        fimg[i] = new float[totalPixels];

	// Read in tiles into fimg
    int px_i = 0;
    for (int y = shrinkBuffers; y < bufferHeight-shrinkBuffers; y++ )
    {
        for (int x = shrinkBuffers; x < bufferWidth-shrinkBuffers ; x++ )
        {
            col = x / tileWidth;     // Tile index
            row = y / tileHeight;    // Tile index
            int tx = x % tileWidth;  // Pixel in tile
            int ty = y % tileHeight; // Pixel in tile

            uint16 *pIn = srcData[col][row];  // Read in from a valid tile
            if (pIn != NULL)  // Our tile
            {
                if (tx >= activeWidth [col][row])
                	tx = max(0,  2*(activeWidth [col][row]-1)-tx); // Fixes tiles with invalid pixels after the padding
                if (ty >= activeHeight[col][row])
                	ty = max(0,  2*(activeHeight[col][row]-1)-ty);
                pIn += (rowWords[col][row] * (padh+ty))   +   (padw+tx);
            }
            
            // Handle non existing tiles
            
            // Top left corner, 180° rotate
            else  if (row < NEIGHBORTILES && col < NEIGHBORTILES)
            {
                pIn = srcData[col+1][row+1];
                if (pIn != NULL)
                    pIn += (rowWords[col+1][row+1] * (padh+((tileHeight-1)-ty))) + (padw+((tileWidth-1)-tx));
            }
            // Top row mirroring
            else if (row < NEIGHBORTILES)
            {
                pIn = srcData[col][row+1];
                if (pIn != NULL)
                    pIn += (rowWords[col][row+1] * (padh+((tileHeight-1)-ty))) + (padw+tx);
            }
            
            // Left col mirroring
            else  if (col < NEIGHBORTILES)
            {
                pIn = srcData[col+1][row];
                if (pIn != NULL)
                    pIn += (rowWords[col+1][row] * (padh+ty)) + (padw+((tileWidth-1)-tx));
            }
            
            // Bottom row mirroring
            else  if (row > NEIGHBORTILES)
            {
                pIn = srcData[col][row-1];
                if (pIn != NULL)
                    pIn += (rowWords[col][row-1] * (padh+((tileHeight-1)-ty))) + (padw+tx);
            }
            
            // Right col mirroring
            else if (col > NEIGHBORTILES)
            {
                pIn = srcData[col-1][row];
                if (pIn != NULL)
                    pIn += (rowWords[col-1][row] * (padh+ty)) + (padw+((tileWidth-1)-tx));
            }
            
            // MISSING: bottom left, bottom right, top right

            // No source tile
            else
                pIn = NULL;

            // Fill the fimg array
            if (pIn != NULL)
            {
                fimg[0][px_i] = INT16_TO_FLOAT(*(pIn));
                fimg[1][px_i] = INT16_TO_FLOAT(*(pIn +  pw)) ;
                fimg[2][px_i] = INT16_TO_FLOAT(*(pIn + 2*pw));
            }
   			
   			// Give it a pretty average init :D
			else
            {
                fimg[0][px_i] = INT16_TO_FLOAT(0x7fff);
                fimg[1][px_i] = INT16_TO_FLOAT(0x7fff);
                fimg[2][px_i] = INT16_TO_FLOAT(0x7fff);

            }

            // Final step, increment pointer
			px_i++;
        }
    }

	// Process the fimg buffer
	ProcessBuffer(fimg,bufferWidth-2*shrinkBuffers,bufferHeight-2*shrinkBuffers,zoomLevel,options,settings,tile,layerOptions,layerPos);
	
    // Write data to processed tile
    px_i = 0;
    uint16 *pOut = targetImage->data16();
    for (int y = -padh; y < tileHeight+pad; y++ )
    {
        for (int x = -padw; x < tileWidth+pad ; x++ )
        {
            int fx = x + (NEIGHBORTILES*tileWidth)-shrinkBuffers;  // Pixel in fimg
            int fy = y + (NEIGHBORTILES*tileHeight)-shrinkBuffers; // Pixel in fimg

            px_i = fy*(numTiles*tileWidth-2*shrinkBuffers) + fx; // Skip unneeded

            // Color components (can be RGB or LAB, depending on the filter position)
            *(pOut)         = FLOAT_TO_INT16(fimg[0][px_i]); // R
            *(pOut + pw)    = FLOAT_TO_INT16(fimg[1][px_i]); // G
            *(pOut + 2*pw)  = FLOAT_TO_INT16(fimg[2][px_i]); // B

            pOut += np;
        }
    }

	// Free memory buffer 
    for (int i = 0; i < 3; i++)
        delete[] fimg[i];
}
