/*
 * (C) 2012, Roland Baudin <roland65@free.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

// Plugin includes
#include "USMPlugin.h"
#include "USMFilter.h"

// Bibble includes
#include "PluginHub.h"
#include "PluginDependency.h"

// Qt includes
#include <QDebug>
#include <QMessageBox>
#include <QSettings>

// C++ includes
#include <iostream>


#define RADIUS_DEFAULT	1.0
#define AMOUNT_DEFAULT	100.0
#define THRESHOLD_DEFAULT 0


#define PLUGIN_NAME_HR "USM"

extern "C" BIBBLE_API BaseB5Plugin *b5plugin()
{
    return new USMPlugin;  // Give the class name, here
}


// Init function
bool USMPlugin::init(PluginHub *hub, int id, int groupId, const QString &)
{
    m_hub = hub;
    m_id = id;
    m_groupId = groupId;

	// Store our group ID and group name (andrewj)
    QSettings oSettings("Bibble and AfterShot Plugins", "PluginGroups");
    oSettings.setValue(group(), groupId);

    return true;
}


// Register the filter
// USM filter must be set just after RgbToLab because it is run on the LAB L channel
bool USMPlugin::registerFilters()
{
    USMFilter *usm = new USMFilter(m_hub, m_groupId);
    if (m_hub->addFilter(usm, PluginHub::Before, QString("Sharpen")))
    {
        qDebug() << "Successfully registered "<<PLUGIN_NAME_HR<<" filter";
        return true;
    }
    qDebug() << "Failed to register "<<PLUGIN_NAME_HR<<" filter";
    return false;
}


// Register the options
bool USMPlugin::registerOptions()
{
    m_hub->addBoolOption(USMFilter::Enable, "enable" , tr("USM enabled"), tr("Enable USM plugin"), "USMFilter", false, false, 0);
    m_hub->addDoubleOption(USMFilter::Radius, "radius" , tr("USM radius"), tr("Radius of the blurring filter (pixels)"), "USMFilter", RADIUS_DEFAULT, 0, 0);
    m_hub->addDoubleOption(USMFilter::Amount, "amount" , tr("USM amount"), tr("Amount of sharpening (percent)"), "USMFilter", AMOUNT_DEFAULT, 0, 0);
    m_hub->addIntOption(USMFilter::Threshold, "threshold" , tr("USM threshold"), tr("Minimal sharpening distance (percent)"), "USMFilter", THRESHOLD_DEFAULT, 0, 0);

    return true;
}


// This function is called after registerFilters()
bool USMPlugin::finish()
{
	// Connect to controlChanged so we can do some settings dependency logic
	connect(m_hub,
	        SIGNAL(controlChanged(const QString &, int, int , const PluginImageSettings &, const PluginOptionList &, PluginOptionList &)),
	        this,
	        SLOT(handleControlChange(const QString &, int, int , const PluginImageSettings &, const PluginOptionList &, PluginOptionList &)));

    return true;
}


PluginDependency *USMPlugin::createDependency(const QString &name)
{
    Q_UNUSED(name);
    return NULL;
}


QList<QString> USMPlugin::toolFiles()
{
    return QList<QString>();
}


QList<QWidget*> USMPlugin::toolWidgets()
{
    return QList<QWidget*>();
}


void USMPlugin::toolWidgetCreated(QWidget *uiWidget)
{
	// Connect the about button to its slot 
	aboutButton = uiWidget->findChild<QPushButton*>("about_button");
	connect(aboutButton, SIGNAL(clicked()), this, SLOT(showAbout()));

	// Connect the reset button to its slot 
	resetButton = uiWidget->findChild<QPushButton*>("reset_button");
	connect(resetButton, SIGNAL(clicked( bool )), this, SLOT(resetClicked(bool)));

	// Connect the enable button to its slot 
	enableCheckBox = uiWidget->findChild<QCheckBox*>("enable_checkbox");

	// Connect the radius edit button to its slot 
	radiusEdit= uiWidget->findChild<QDoubleSpinBox*>("radius_edit");

	// Connect the amount edit button to its slot 
	amountEdit= uiWidget->findChild<QDoubleSpinBox*>("amount_edit");

	// Connect the threshol edit button to its slot 
	thresholdEdit= uiWidget->findChild<QSpinBox*>("threshold_edit");
}


// Enable the filter when the sliders change
void USMPlugin::handleControlChange(const QString &optionName, int groupId, int layer, const PluginImageSettings &options, const PluginOptionList &current, PluginOptionList &changes)
{
	Q_UNUSED( current );
	
	// Check if one of the slider settings was changed
	if (groupId == m_groupId && ( optionName == "radius" || optionName == "amount" || optionName == "threshold"))
	{
		// A slider did change - Set "enable" to true if it isn't
		bool okay;
		bool enable = options.options(layer)->getBool(USMFilter::Enable, m_groupId, okay);
		if (enable == false)
			changes.setBool(USMFilter::Enable, m_groupId, "true");
	}
}


// Display an info message in a window
void USMPlugin::showAbout()
{
	QMessageBox::about(NULL, tr("USM Plugin"),
tr("USM Plugin version 0.6\n\
================\n\n\
Here is the standard Unsharp Mask plugin for picture sharpening and local contrast enhancement.\n\n\
The unsharp mask works by utilizing a slightly blurred version of the original image, which is then subtracted away from the original to detect the presence of edges.\n\n\
It can also be used (with large radius and small amount) for local constrast enhancement.\n\
Note the filtering is done on the Lab L channel.\n\n\
Typical parameter values:\n\
- Sharpening: radius = 1, amount = 100, threshold = 0\n\
- Local contrast enhancement: radius = 60, amount = 20, threshold = 0\n\n\
Tip: both effects can be combined by using two separate layers.\n\n\
Caution: very large radius values (>100) can sometimes cause tiling image artefacts.\n\n\
USM 0.6\n\
Copyright (C) 2012 Roland Baudin\n\
http://roland65.free.fr"));
}



// Reset plugin
void USMPlugin::resetClicked(bool)
{
	disconnect(m_hub, SIGNAL(controlChanged(const QString &, int, int , const PluginImageSettings &, const PluginOptionList &, PluginOptionList &)), this, SLOT(handleControlChange(const QString &, int, int , const PluginImageSettings &, const PluginOptionList &, PluginOptionList &)));
	enableCheckBox->setChecked(false);
	radiusEdit->setValue(RADIUS_DEFAULT);
	amountEdit->setValue(AMOUNT_DEFAULT);
	thresholdEdit->setValue(THRESHOLD_DEFAULT);
	connect(m_hub, SIGNAL(controlChanged(const QString &, int, int , const PluginImageSettings &, const PluginOptionList &, PluginOptionList &)), this, SLOT(handleControlChange(const QString &, int, int , const PluginImageSettings &, const PluginOptionList &, PluginOptionList &)));
}
